# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ruler widget for the PyAnsys Geometry plotter."""

import os

from vtk import vtkButtonWidget, vtkPNGReader

from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.plotting.widgets.widget import PlotterWidget


class ShowDesignPoints(PlotterWidget):
    """
    Provides the a button to hide/show DesignPoint objects in the plotter.

    Parameters
    ----------
    plotter_helper : PlotterHelper
        Provides the plotter to add the button to.
    """

    def __init__(self, plotter_helper: "PlotterHelper") -> None:
        """Initialize the ``ShowDesignPoints`` class."""
        # Call PlotterWidget ctor
        super().__init__(plotter_helper._pl.scene)
        self.plotter_helper = plotter_helper

        # Initialize variables
        self._geom_object_actors_map = self.plotter_helper._pl._geom_object_actors_map
        self._button: vtkButtonWidget = self.plotter_helper._pl.scene.add_checkbox_button_widget(
            self.callback, position=(5, 438), size=30, border_size=3
        )

    def callback(self, state: bool) -> None:
        """
        Remove or add the DesignPoint actors upon click.

        Parameters
        ----------
        state : bool
            State of the button, which is inherited from PyVista. The value is ``True``
            if the button is active.
        """
        if not state:
            for actor, object in self._geom_object_actors_map.items():
                if isinstance(object, DesignPoint):
                    self.plotter_helper._pl.scene.add_actor(actor)
        else:
            for actor, object in self._geom_object_actors_map.items():
                if isinstance(object, DesignPoint):
                    self.plotter_helper._pl.scene.remove_actor(actor)

    def update(self) -> None:
        """Define the configuration and representation of the button widget button."""
        show_point_vr = self._button.GetRepresentation()
        show_point_icon_file = os.path.join(os.path.dirname(__file__), "_images", "designpoint.png")
        show_point_r = vtkPNGReader()
        show_point_r.SetFileName(show_point_icon_file)
        show_point_r.Update()
        image = show_point_r.GetOutput()
        show_point_vr.SetButtonTexture(0, image)
        show_point_vr.SetButtonTexture(1, image)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the ``ProductInstance`` class."""
import os
import signal
import socket
import subprocess

from ansys.tools.path import get_available_ansys_installations, get_latest_ansys_installation
from beartype.typing import TYPE_CHECKING, Dict, List

from ansys.geometry.core.connection.backend import ApiVersions, BackendType
from ansys.geometry.core.logger import LOG

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.modeler import Modeler


WINDOWS_GEOMETRY_SERVICE_FOLDER = "GeometryService"
"""Default Geometry Service's folder name into the unified installer."""

DISCOVERY_FOLDER = "Discovery"
"""Default Discovery's folder name into the unified installer."""

SPACECLAIM_FOLDER = "scdm"
"""Default SpaceClaim's folder name into the unified installer."""

ADDINS_SUBFOLDER = "Addins"
"""Default global Addins's folder name into the unified installer."""

BACKEND_SUBFOLDER = "ApiServer"
"""Default backend's folder name into the ``ADDINS_SUBFOLDER`` folder."""

MANIFEST_FILENAME = "Presentation.ApiServerAddIn.Manifest.xml"
"""
Default backend's addin filename.

To be used only for local start of Ansys Discovery or Ansys SpaceClaim.
"""

GEOMETRY_SERVICE_EXE = "Presentation.ApiServerDMS.exe"
"""The Windows Geometry Service's filename."""

DISCOVERY_EXE = "Discovery.exe"
"""The Ansys Discovery's filename."""

SPACECLAIM_EXE = "SpaceClaim.exe"
"""The Ansys SpaceClaim's filename."""

BACKEND_LOG_LEVEL_VARIABLE = "LOG_LEVEL"
"""The backend's log level environment variable for local start."""

BACKEND_TRACE_VARIABLE = "ENABLE_TRACE"
"""The backend's enable trace environment variable for local start."""

BACKEND_HOST_VARIABLE = "API_ADDRESS"
"""The backend's ip address environment variable for local start."""

BACKEND_PORT_VARIABLE = "API_PORT"
"""The backend's port number environment variable for local start."""

BACKEND_LOGS_FOLDER_VARIABLE = "ANS_DSCO_REMOTE_LOGS_FOLDER"
"""The backend's logs folder path to be used."""

BACKEND_API_VERSION_VARIABLE = "API_VERSION"
"""
The backend's api version environment variable for local start.

To be used only with Ansys Discovery and Ansys SpaceClaim.
"""

BACKEND_SPACECLAIM_OPTIONS = "--spaceclaim-options"
"""
The additional argument for local Ansys Discovery start.

To be used only with Ansys Discovery.
"""

BACKEND_ADDIN_MANIFEST_ARGUMENT = "/ADDINMANIFESTFILE="
"""
The argument to specify the backend's addin manifest file's path.

To be used only with Ansys Discovery and Ansys SpaceClaim.
"""

BACKEND_SPACECLAIM_HIDDEN = "/Headless=True"
"""
The argument to hide SpaceClaim's UI on the backend.

To be used only with Ansys SpaceClaim.
"""

BACKEND_DISCOVERY_HIDDEN = "--hidden"
"""
The argument to hide Discovery's UI on the backend.

To be used only with Ansys Discovery.
"""

BACKEND_SPLASH_OFF = "/Splash=False"
"""
The argument to specify the backend's addin manifest file's path.

To be used only with Ansys Discovery and Ansys SpaceClaim.
"""


class ProductInstance:
    """
    ``ProductInstance`` class.

    This class is used as a handle for a local session of Ansys Product's backend: Discovery,
    Windows Geometry Service or SpaceClaim.

    Parameters
    ----------
    pid : int
        The local instance's process identifier. This allows to keep track of the process
        and close it if need be.
    """

    def __init__(self, pid: int):
        """Initialize ``ProductInstance`` class."""
        self._pid = pid

    def close(self) -> bool:
        """Close the process associated to the pid."""
        try:
            os.kill(self._pid, signal.SIGTERM)
        except OSError as oserr:
            LOG.error(str(oserr))
            return False

        return True


def prepare_and_start_backend(
    backend_type: BackendType,
    product_version: int = None,
    host: str = "localhost",
    port: int = None,
    enable_trace: bool = False,
    log_level: int = 2,
    api_version: ApiVersions = ApiVersions.LATEST,
    timeout: int = 150,
    manifest_path: str = None,
    logs_folder: str = None,
    hidden: bool = False,
) -> "Modeler":
    """
    Start the requested service locally using the ``ProductInstance`` class.

    When calling this method, a standalone service or product session is started.
    By default, if an endpoint is specified (by defining `host` and `port` parameters)
    but the endpoint is not available, the startup will fail. Otherwise, it will try to
    launch its own service.

    Parameters
    ----------
    product_version: ``int``, optional
        The product version to be started. Goes from v23.2.1 to
        the latest. Default is ``None``.
        If a specific product version is requested but not installed locally,
        a SystemError will be raised.
    host: str, optional
        IP address at which the Geometry service will be deployed. By default,
        its value will be ``localhost``.
    port : int, optional
        Port at which the Geometry service will be deployed. By default, its
        value will be ``None``.
    enable_trace : bool, optional
        Boolean enabling the logs trace on the Geometry service console window.
        By default its value is ``False``.
    log_level : int, optional
        Backend's log level from 0 to 3:
            0: Chatterbox
            1: Debug
            2: Warning
            3: Error

        The default is ``2`` (Warning).
    api_version: ``ApiVersions``, optional
        The backend's API version to be used at runtime. Goes from API v21 to
        the latest. Default is ``ApiVersions.LATEST``.
    timeout : int, optional
        Timeout for starting the backend startup process. The default is 150.
    manifest_path : str, optional
        Used to specify a manifest file path for the ApiServerAddin. This way,
        it is possible to run an ApiServerAddin from a version an older product
        version. Only applicable for Ansys Discovery and Ansys SpaceClaim.
    logs_folder : sets the backend's logs folder path. If nothing is defined,
        the backend will use its default path.
    hidden : starts the product hiding its UI. Default is ``False``.

    Raises
    ------
    ConnectionError
        If the specified endpoint is already in use, a connection error will be raised.
    SystemError
        If there is not an Ansys product 23.2 version or later installed
        or if a specific product's version is requested but not installed locally then
        a SystemError will be raised.

    Returns
    -------
    Modeler
        Instance of the Geometry service.
    """
    from ansys.geometry.core.modeler import Modeler

    port = _check_port_or_get_one(port)
    installations = get_available_ansys_installations()
    if product_version != None:
        _check_version_is_available(product_version, installations)
    else:
        product_version = get_latest_ansys_installation()[0]
        _check_minimal_versions(product_version)

    args = []
    env_copy = _get_common_env(
        host=host,
        port=port,
        enable_trace=enable_trace,
        log_level=log_level,
        logs_folder=logs_folder,
    )

    if backend_type == BackendType.DISCOVERY:
        args.append(os.path.join(installations[product_version], DISCOVERY_FOLDER, DISCOVERY_EXE))
        if hidden is True:
            args.append(BACKEND_DISCOVERY_HIDDEN)

        # Here begins the spaceclaim arguments.
        args.append(BACKEND_SPACECLAIM_OPTIONS)
        args.append(
            BACKEND_ADDIN_MANIFEST_ARGUMENT
            + _manifest_path_provider(product_version, installations, manifest_path)
        )
        env_copy[BACKEND_API_VERSION_VARIABLE] = str(api_version)
    elif backend_type == BackendType.SPACECLAIM:
        args.append(os.path.join(installations[product_version], SPACECLAIM_FOLDER, SPACECLAIM_EXE))
        if hidden is True:
            args.append(BACKEND_SPACECLAIM_HIDDEN)
            args.append(BACKEND_SPLASH_OFF)
        args.append(
            BACKEND_ADDIN_MANIFEST_ARGUMENT
            + _manifest_path_provider(product_version, installations, manifest_path)
        )
        env_copy[BACKEND_API_VERSION_VARIABLE] = str(api_version)
    elif backend_type == BackendType.WINDOWS_SERVICE:
        latest_version = get_latest_ansys_installation()[0]
        args.append(
            os.path.join(
                installations[latest_version], WINDOWS_GEOMETRY_SERVICE_FOLDER, GEOMETRY_SERVICE_EXE
            )
        )
    else:
        raise RuntimeError(
            f"Cannot connect to backend {backend_type.name} using ``prepare_and_start_backend()``"
        )

    LOG.info(f"Launching ProductInstance for {backend_type.name}")
    LOG.debug(f"Args: {args}")
    LOG.debug(f"Environment variables: {env_copy}")

    instance = ProductInstance(_start_program(args, env_copy).pid)

    # Verify that the backend is ready to accept connections
    # before returning the Modeler instance.
    LOG.info("Waiting for backend to be ready...")
    _wait_for_backend(host, port, timeout)

    return Modeler(
        host=host, port=port, timeout=timeout, product_instance=instance, backend_type=backend_type
    )


def get_available_port() -> int:
    """
    Return an available port to be used.

    Returns
    -------
    int
        The available port.
    """
    sock = socket.socket()
    sock.bind((socket.gethostname(), 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def _wait_for_backend(host: str, port: int, timeout: int):
    """
    Check if the backend is ready to accept connections.

    Parameters
    ----------
    host : str
        The backend's ip address.
    port : int
        The backend's port number.
    timeout : int
        The timeout in seconds.
    """
    import time

    start_time = time.time()
    while time.time() - start_time < timeout:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex((host, port)) == 0:
                LOG.debug("Backend is ready to accept connections.")
                return
            else:
                LOG.debug("Still waiting for backend to be ready... Retrying in 5 seconds.")
                time.sleep(5)

    raise ConnectionError("Timeout while waiting for backend to be ready.")


def _is_port_available(port: int, host: str = "localhost") -> bool:
    """
    Check whether the argument port is available or not.

    The optional argument is the ip address where to check port availability.
    Its default is ``localhost``.
    """
    if port != 0:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.bind((host, port))
                return True
            except:
                return False


def _manifest_path_provider(
    version: int, available_installations: Dict, manifest_path: str = None
) -> str:
    """Return the ApiServer's addin manifest file path."""
    if manifest_path:
        if os.path.exists(manifest_path):
            return manifest_path
        else:
            LOG.warning(
                "Specified manifest file's path does not exist. Taking install default path."
            )

    # Default manifest path
    def_manifest_path = os.path.join(
        available_installations[version], ADDINS_SUBFOLDER, BACKEND_SUBFOLDER, MANIFEST_FILENAME
    )

    if os.path.exists(def_manifest_path):
        return def_manifest_path
    else:
        msg = (
            "Default manifest file's path does not exist."
            " Please specify a valid manifest."
            f" The ApiServer Add-In seems to be missing in {os.path.dirname(def_manifest_path)}"
        )
        LOG.error(msg)
        raise RuntimeError(msg)


def _start_program(args: List[str], local_env: Dict[str, str]) -> subprocess.Popen:
    """
    Start the program where the path is the first item of the ``args`` array argument.

    Parameters
    ----------
    args : List[str]
        List of arguments to be passed to the program. The first list's item shall
        be the program path.
    local_env : Dict[str,str]
        Environment variables to be passed to the program.

    Returns
    -------
    subprocess.Popen
        The subprocess object.
    """
    return subprocess.Popen(
        args,
        shell=os.name != "nt",
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        env=local_env,
    )


def _check_minimal_versions(latest_installed_version: int) -> None:
    """
    Check client is compatible with Ansys Products starting from 2023.2.1 version.

    Check that at least V232 is installed.
    """
    if abs(latest_installed_version) < 232:
        msg = (
            "PyAnsys Geometry is compatible with Ansys Products from version 23.2.1. "
            + "Please install Ansys products 23.2.1 or later."
        )
        raise SystemError(msg)


def _check_version_is_available(version: int, installations: Dict[int, str]) -> None:
    """Check that the requested version for launcher is installed on the system."""
    if version not in installations:
        msg = (
            f"The requested Ansys product's version {version} is not available, "
            + "please specify a different version."
        )
        raise SystemError(msg)


def _check_port_or_get_one(port: int) -> int:
    """
    If a ``port`` argument is specified, check that it's free.

    If not, raise an error.

    If ``port`` is None, return an available port by calling ``get_available_port``.
    """
    if port:
        if _is_port_available(port):
            return port
        else:
            msg = f"Port {port} is already in use. Please specify a different one."
            raise ConnectionError(msg)
    else:
        return get_available_port()


def _get_common_env(
    host: str,
    port: int,
    enable_trace: bool,
    log_level: int,
    logs_folder: str = None,
) -> Dict[str, str]:
    """
    Make a copy of the actual system's environment.

    Then update or create some environment variables with the provided arguments.
    """
    env_copy = os.environ.copy()

    env_copy[BACKEND_HOST_VARIABLE] = host
    env_copy[BACKEND_PORT_VARIABLE] = f"{port}"
    env_copy[BACKEND_TRACE_VARIABLE] = f"{enable_trace:d}"
    env_copy[BACKEND_LOG_LEVEL_VARIABLE] = f"{log_level}"
    if logs_folder is not None:
        env_copy[BACKEND_LOGS_FOLDER_VARIABLE] = f"{logs_folder}"

    return env_copy



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a sketch."""

from beartype import beartype as check_input_types
from beartype.typing import TYPE_CHECKING, Dict, List, Optional, Union
from pint import Quantity

from ansys.geometry.core.math.constants import ZERO_POINT2D
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.math.vector import UnitVector3D, Vector2D, Vector3D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.sketch.arc import Arc
from ansys.geometry.core.sketch.box import Box
from ansys.geometry.core.sketch.circle import SketchCircle
from ansys.geometry.core.sketch.edge import SketchEdge
from ansys.geometry.core.sketch.ellipse import SketchEllipse
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.gears import DummyGear, SpurGear
from ansys.geometry.core.sketch.polygon import Polygon
from ansys.geometry.core.sketch.segment import SketchSegment
from ansys.geometry.core.sketch.slot import Slot
from ansys.geometry.core.sketch.trapezoid import Trapezoid
from ansys.geometry.core.sketch.triangle import Triangle
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from pyvista import PolyData

SketchObject = Union[SketchEdge, SketchFace]
"""Type used to refer to both ``SketchEdge`` and ``SketchFace`` as possible values."""


class Sketch:
    """Provides for building 2D sketch elements."""

    # Types of the class instance private attributes
    _faces: List[SketchFace]
    _edges: List[SketchEdge]
    _current_sketch_context: List[SketchObject]
    _tags: Dict[str, List[SketchObject]]

    @check_input_types
    def __init__(
        self,
        plane: Optional[Plane] = Plane(),
    ):
        """Initialize the ``Sketch`` class."""
        self._plane = plane

        self._faces = []
        self._edges = []
        self._current_sketch_context = []
        self._sketches_polydata_selection = []

        # data structure to track tagging individual
        # sketch objects and collections of sketch objects
        self._tags = {}

    @property
    def plane(self) -> Plane:
        """Sketch plane configuration."""
        return self._plane

    @property
    def edges(self) -> List[SketchEdge]:
        """
        List of all independently sketched edges.

        Notes
        -----
        Independently sketched edges are not assigned to a face. Face edges
        are not included in this list.
        """
        return self._edges

    @property
    def faces(self) -> List[SketchFace]:
        """List of all independently sketched faces."""
        return self._faces

    @plane.setter
    @check_input_types
    def plane(self, plane: Plane) -> None:
        """
        Set the sketch plane configuration.

        Parameters
        ----------
        plane : Plane
            New plane for the sketch planar orientation.
        """
        self._plane = plane
        [face.plane_change(plane) for face in self.faces]
        [edge.plane_change(plane) for edge in self.edges]

    @check_input_types
    def translate_sketch_plane(self, translation: Vector3D) -> "Sketch":
        """
        Translate the origin location of the active sketch plane.

        Parameters
        ----------
        translation : Vector3D
            Vector defining the translation. Meters is the expected unit.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        self.plane = Plane(
            self.plane.origin + translation, self.plane.direction_x, self.plane.direction_y
        )
        return self

    @check_input_types
    def translate_sketch_plane_by_offset(
        self,
        x: Union[Quantity, Distance] = Quantity(0, DEFAULT_UNITS.LENGTH),
        y: Union[Quantity, Distance] = Quantity(0, DEFAULT_UNITS.LENGTH),
        z: Union[Quantity, Distance] = Quantity(0, DEFAULT_UNITS.LENGTH),
    ) -> "Sketch":
        """
        Translate the origin location of the active sketch plane by offsets.

        Parameters
        ----------
        x : Union[~pint.Quantity, Distance], default: ~pint.Quantity(0, ``DEFAULT_UNITS.LENGTH``)
            Amount to translate the origin of the x-direction.
        y : Union[~pint.Quantity, Distance], default: ~pint.Quantity(0, ``DEFAULT_UNITS.LENGTH``)
            Amount to translate the origin of the y-direction.
        z : Union[~pint.Quantity, Distance], default: ~pint.Quantity(0, ``DEFAULT_UNITS.LENGTH``)
            Amount to translate the origin of the z-direction.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        x_magnitude = (
            x.m_as(DEFAULT_UNITS.LENGTH)
            if not isinstance(x, Distance)
            else x.value.m_as(DEFAULT_UNITS.LENGTH)
        )

        y_magnitude = (
            y.m_as(DEFAULT_UNITS.LENGTH)
            if not isinstance(y, Distance)
            else y.value.m_as(DEFAULT_UNITS.LENGTH)
        )

        z_magnitude = (
            z.m_as(DEFAULT_UNITS.LENGTH)
            if not isinstance(z, Distance)
            else z.value.m_as(DEFAULT_UNITS.LENGTH)
        )
        translation = Vector3D([x_magnitude, y_magnitude, z_magnitude])
        return self.translate_sketch_plane(translation)

    @check_input_types
    def translate_sketch_plane_by_distance(
        self, direction: UnitVector3D, distance: Union[Quantity, Distance]
    ) -> "Sketch":
        """
        Translate the origin location active sketch plane by distance.

        Parameters
        ----------
        direction : UnitVector3D
            Direction to translate the origin.
        distance : Union[~pint.Quantity, Distance]
            Distance to translate the origin.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        magnitude = (
            distance.m_as(DEFAULT_UNITS.LENGTH)
            if not isinstance(distance, Distance)
            else distance.value.m_as(DEFAULT_UNITS.LENGTH)
        )
        translation = Vector3D(
            [direction.x * magnitude, direction.y * magnitude, direction.z * magnitude]
        )
        return self.translate_sketch_plane(translation)

    @check_input_types
    def get(self, tag: str) -> List[SketchObject]:
        """
        Get a list of shapes with a given tag.

        Parameters
        ----------
        tag : str
            Tag to query against.
        """
        return self._tags[tag]

    @check_input_types
    def face(self, face: SketchFace, tag: Optional[str] = None) -> "Sketch":
        """
        Add a sketch face to the sketch.

        Parameters
        ----------
        face : SketchFace
            Face to add.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        self._faces.append(face)
        if tag:
            self._tag([face], tag)

        return self

    @check_input_types
    def edge(self, edge: SketchEdge, tag: Optional[str] = None) -> "Sketch":
        """
        Add a sketch edge to the sketch.

        Parameters
        ----------
        edge : SketchEdge
            Edge to add.
        tag : str, default: None
            User-defined label for identifying the edge.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        self._edges.append(edge)
        if tag:
            self._tag([edge], tag)

        return self

    @check_input_types
    def select(self, *tags: str) -> "Sketch":
        """Add all objects that match provided tags to the current context."""
        self._current_sketch_context = []

        for tag in tags:
            self._current_sketch_context.extend(self._tags[tag])

        return self

    def segment(self, start: Point2D, end: Point2D, tag: Optional[str] = None) -> "Sketch":
        """
        Add a segment sketch object to the sketch plane.

        Parameters
        ----------
        start : Point2D
            Starting point of the line segment.
        end : Point2D
            Ending point of the line segment.
        tag : str, default: None
            User-defined label for identifying the edge.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        segment = SketchSegment(start, end)
        return self.edge(segment, tag)

    def segment_to_point(self, end: Point2D, tag: Optional[str] = None) -> "Sketch":
        """
        Add a segment to the sketch plane starting from the previous edge end point.

        Parameters
        ----------
        end : Point2D
            Ending point of the line segment.
        tag : str, default: None
            User-defined label for identifying the edge.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.

        Notes
        -----
        The starting point of the created edge is based upon the current context
        of the sketch, such as the end point of a previously added edge.
        """
        segment = SketchSegment(self._single_point_context_reference(), end)
        return self.edge(segment, tag)

    @check_input_types
    def segment_from_point_and_vector(
        self, start: Point2D, vector: Vector2D, tag: Optional[str] = None
    ):
        """
        Add a segment to the sketch starting from a given starting point.

        Notes
        -----
        Vector magnitude determines the segment endpoint.
        Vector magnitude is assumed to use the same unit as the starting point.

        Parameters
        ----------
        start : Point2D
            Starting point of the line segment.
        vector : Vector2D
            Vector defining the line segment. Vector magnitude determines
            the segment endpoint. Vector magnitude is assumed to be in the
            same unit as the starting point.
        tag : str, default: None
            User-defined label for identifying the edge.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        end_vec_as_point = Point2D(vector, start.unit)
        end = start + end_vec_as_point

        return self.segment(start, end, tag)

    @check_input_types
    def segment_from_vector(self, vector: Vector2D, tag: Optional[str] = None):
        """
        Add a segment to the sketch starting from the end point of the previous edge.

        Parameters
        ----------
        vector : Vector2D
            Vector defining the line segment.
        tag : str, default: None
            User-defined label for identifying the edge.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.

        Notes
        -----
        The starting point of the created edge is based upon the current context
        of the sketch, such as the end point of a previously added edge.

        Vector magnitude determines the segment endpoint.
        Vector magnitude is assumed to use the same unit as the starting point
        in the previous context.
        """
        start = self._single_point_context_reference()
        return self.segment_from_point_and_vector(start, vector, tag)

    def arc(
        self,
        start: Point2D,
        end: Point2D,
        center: Point2D,
        clockwise: Optional[bool] = False,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Add an arc to the sketch plane.

        Parameters
        ----------
        start : Point2D
            Starting point of the arc.
        end : Point2D
            Ending point of the arc.
        center : Point2D
            Center point of the arc.
        clockwise : bool, default: False
            Whether the arc spans the angle clockwise between the start
            and end points. When ``False `` (default), the arc spans the angle
            counter-clockwise. When ``True``, the arc spans the angle
            clockwise.
        tag : str, default: None
            User-defined label for identifying the edge.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        arc = Arc(center, start, end, clockwise)
        return self.edge(arc, tag)

    @check_input_types
    def arc_to_point(
        self,
        end: Point2D,
        center: Point2D,
        clockwise: Optional[bool] = False,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Add an arc to the sketch starting from the end point of the previous edge.

        Parameters
        ----------
        end : Point2D
            Ending point of the arc.
        center : Point2D
            Center point of the arc.
        clockwise : bool, default: False
            Whether the arc spans the angle clockwise between the start
            and end points. When ``False`` (default), the arc spans the angle
            counter-clockwise. When ``True``, the arc spans the angle
            clockwise.
        tag : str, default: None
            User-defined label for identifying the edge.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.

        Notes
        -----
        The starting point of the created edge is based upon the current context
        of the sketch, such as the end point of a previously added edge.
        """
        start = self._single_point_context_reference()
        arc = Arc(center, start, end, clockwise)
        return self.edge(arc, tag)

    def arc_from_three_points(
        self,
        start: Point2D,
        inter: Point2D,
        end: Point2D,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Add an arc to the sketch plane from three given points.

        Parameters
        ----------
        start : Point2D
            Starting point of the arc.
        inter : Point2D
            Intermediate point (location) of the arc.
        end : Point2D
            End point of the arc.
        tag : str, default: None
            User-defined label for identifying the edge.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        arc = Arc.from_three_points(start, inter, end)
        return self.edge(arc, tag)

    def triangle(
        self,
        point1: Point2D,
        point2: Point2D,
        point3: Point2D,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Add a triangle to the sketch using given vertex points.

        Parameters
        ----------
        point1 : Point2D
            Point that represents a vertex of the triangle.
        point2 : Point2D
            Point that represents a vertex of the triangle.
        point3 : Point2D
            Point that represents a vertex of the triangle.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        triangle = Triangle(point1, point2, point3)
        return self.face(triangle, tag)

    def trapezoid(
        self,
        width: Union[Quantity, Distance, Real],
        height: Union[Quantity, Distance, Real],
        slant_angle: Union[Quantity, Angle, Real],
        nonsymmetrical_slant_angle: Optional[Union[Quantity, Angle, Real]] = None,
        center: Optional[Point2D] = ZERO_POINT2D,
        angle: Optional[Union[Quantity, Angle, Real]] = 0,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Add a triangle to the sketch using given vertex points.

        Parameters
        ----------
        width : Union[~pint.Quantity, Distance, Real]
            Width of the slot main body.
        height : Union[~pint.Quantity, Distance, Real]
            Height of the slot.
        slant_angle : Union[~pint.Quantity, Angle, Real]
            Angle for trapezoid generation.
        nonsymmetrical_slant_angle : Union[~pint.Quantity, Angle, Real], default: None
            Asymmetrical slant angles on each side of the trapezoid.
            The default is ``None``, in which case the trapezoid is symmetrical.
        center : Point2D, default: (0, 0)
            Center point of the trapezoid.
        angle : Optional[Union[~pint.Quantity, Angle, Real]], default: 0
            Placement angle for orientation alignment.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        trapezoid = Trapezoid(width, height, slant_angle, nonsymmetrical_slant_angle, center, angle)
        return self.face(trapezoid, tag)

    def circle(
        self,
        center: Point2D,
        radius: Union[Quantity, Distance, Real],
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Add a circle to the plane at a given center.

        Parameters
        ----------
        center: Point2D
            Center point of the circle.
        radius : Union[~pint.Quantity, Distance, Real]
            Radius of the circle.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        circle = SketchCircle(center, radius, plane=self.plane)
        return self.face(circle, tag)

    def box(
        self,
        center: Point2D,
        width: Union[Quantity, Distance, Real],
        height: Union[Quantity, Distance, Real],
        angle: Optional[Union[Quantity, Angle, Real]] = 0,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Create a box on the sketch.

        Parameters
        ----------
        center: Point2D
            Center point of the box.
        width : Union[~pint.Quantity, Distance, Real]
            Width of the box.
        height : Union[~pint.Quantity, Distance, Real]
            Height of the box.
        angle : Union[~pint.Quantity, Real], default: 0
            Placement angle for orientation alignment.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        box = Box(center, width, height, angle)
        return self.face(box, tag)

    def slot(
        self,
        center: Point2D,
        width: Union[Quantity, Distance, Real],
        height: Union[Quantity, Distance, Real],
        angle: Optional[Union[Quantity, Angle, Real]] = 0,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Create a slot on the sketch.

        Parameters
        ----------
        center: Point2D
            Center point of the slot.
        width : Union[~pint.Quantity, Distance, Real]
            Width of the slot.
        height : Union[~pint.Quantity, Distance, Real]
            Height of the slot.
        angle : Union[~pint.Quantity, Angle, Real], default: 0
            Placement angle for orientation alignment.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        slot = Slot(center, width, height, angle)
        return self.face(slot, tag)

    def ellipse(
        self,
        center: Point2D,
        major_radius: Union[Quantity, Distance, Real],
        minor_radius: Union[Quantity, Distance, Real],
        angle: Optional[Union[Quantity, Angle, Real]] = 0,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Create an ellipse on the sketch.

        Parameters
        ----------
        center: Point2D
            Center point of the ellipse.
        major_radius : Union[~pint.Quantity, Distance, Real]
            Semi-major axis of the ellipse.
        minor_radius : Union[~pint.Quantity, Distance, Real]
            Semi-minor axis of the ellipse.
        angle : Union[~pint.Quantity, Angle, Real], default: 0
            Placement angle for orientation alignment.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        ellipse = SketchEllipse(center, major_radius, minor_radius, angle)
        return self.face(ellipse, tag)

    def polygon(
        self,
        center: Point2D,
        inner_radius: Union[Quantity, Distance, Real],
        sides: int,
        angle: Optional[Union[Quantity, Angle, Real]] = 0,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Create a polygon on the sketch.

        Parameters
        ----------
        center: Point2D
            Center point of the polygon.
        inner_radius : Union[~pint.Quantity, Distance, Real]
            Inner radius (apothem) of the polygon.
        sides : int
            Number of sides of the polygon.
        angle : Union[~pint.Quantity, Angle, Real], default: 0
            Placement angle for orientation alignment.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        polygon = Polygon(center, inner_radius, sides, angle)
        return self.face(polygon, tag)

    def dummy_gear(
        self,
        origin: Point2D,
        outer_radius: Union[Quantity, Distance, Real],
        inner_radius: Union[Quantity, Distance, Real],
        n_teeth: int,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Create a dummy gear on the sketch.

        Parameters
        ----------
        origin : Point2D
            Origin of the gear.
        outer_radius : Union[~pint.Quantity, Distance, Real]
            Outer radius of the gear.
        inner_radius : Union[~pint.Quantity, Distance, Real]
            Inner radius of the gear.
        n_teeth : int
            Number of teeth of the gear.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        gear = DummyGear(origin, outer_radius, inner_radius, n_teeth)
        return self.face(gear, tag)

    def spur_gear(
        self,
        origin: Point2D,
        module: Real,
        pressure_angle: Union[Quantity, Angle, Real],
        n_teeth: int,
        tag: Optional[str] = None,
    ) -> "Sketch":
        """
        Create a spur gear on the sketch.

        Parameters
        ----------
        origin : Point2D
            Origin of the spur gear.
        module : Real
            Module of the spur gear. This is also the ratio between the pitch circle
            diameter in millimeters and the number of teeth.
        pressure_angle : Union[~pint.Quantity, Angle, Real]
            Pressure angle of the spur gear.
        n_teeth : int
            Number of teeth of the spur gear.
        tag : str, default: None
            User-defined label for identifying the face.

        Returns
        -------
        Sketch
            Revised sketch state ready for further sketch actions.
        """
        gear = SpurGear(origin, module, pressure_angle, n_teeth)
        return self.face(gear, tag)

    @check_input_types
    def tag(self, tag: str) -> None:
        """
        Add a tag to the active selection of sketch objects.

        Parameters
        ----------
        tag : str
            Tag to assign to the sketch objects.
        """
        self._tags[tag] = self._current_sketch_context

    def _single_point_context_reference(self) -> Point2D:
        """
        Get the last reference point from historical context.

        Notes
        -----
        If no single point context is available, ``ZERO_POINT2D`` is returned.
        """
        if not self._edges or len(self._edges) == 0:
            return ZERO_POINT2D

        return self._edges[-1].end

    def _tag(self, sketch_collection: List[SketchObject], tag: str) -> None:
        """
        Add a tag for a collection of sketch objects.

        Parameters
        ----------
        sketch_collection : List[SketchObject]
            Sketch objects to tag.
        tag : str, default: None
            Tag to assign to these sketch objects.
        """
        self._tags[tag] = sketch_collection

    def plot(
        self,
        view_2d: Optional[bool] = False,
        screenshot: Optional[str] = None,
        use_trame: Optional[bool] = None,
        selected_pd_objects: List["PolyData"] = None,
        **plotting_options: Optional[dict],
    ):
        """
        Plot all objects of the sketch to the scene.

        Parameters
        ----------
        view_2d : bool, default: False
            Whether to represent the plot in a 2D format.
        screenshot : str, optional
            Path for saving a screenshot of the image that is being represented.
        use_trame : bool, default: None
            Whether to enables the use of `trame <https://kitware.github.io/trame/index.html>`_.
            The default is ``None``, in which case the ``USE_TRAME`` global
            setting is used.
        **plotting_options : dict, optional
            Keyword arguments for plotting. For allowable keyword arguments,
            see the :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Show the plot requested - i.e. all polydata in sketch
        from ansys.geometry.core.plotting import PlotterHelper

        if view_2d:
            vector = self.plane.direction_z.tolist()
            viewup = self.plane.direction_y.tolist()
            view_2d_dict = {"vector": vector, "viewup": viewup}
        else:
            view_2d_dict = None

        if selected_pd_objects is not None:
            pl_helper = PlotterHelper(use_trame=use_trame).plot(
                selected_pd_objects,
                screenshot=screenshot,
                view_2d=view_2d_dict,
                opacity=0.7,  # We are passing PD objects directly... apply opacity to all of them
                **plotting_options,
            )
        else:
            pl_helper = PlotterHelper(use_trame=use_trame).plot(
                self,
                screenshot=screenshot,
                view_2d=view_2d_dict,
                **plotting_options,
            )

    def plot_selection(
        self,
        view_2d: Optional[bool] = False,
        screenshot: Optional[str] = None,
        use_trame: Optional[bool] = None,
        **plotting_options: Optional[dict],
    ):
        """
        Plot the current selection to the scene.

        Parameters
        ----------
        view_2d : bool, default: False
            Whether to represent the plot in a 2D format.
        screenshot : str, optional
            Path for saving a screenshot of the image that is being represented.
        use_trame : bool, default: None
            Whether to enables the use of `trame <https://kitware.github.io/trame/index.html>`_.
            The default is ``None``, in which case the ``USE_TRAME`` global
            setting is used.
        **plotting_options : dict, optional
            Keyword arguments for plotting. For allowable keyword arguments,
            see the :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Get the selected polydata
        sketches_polydata_selection = []
        sketches_polydata_selection.extend(
            [
                sketch_item.visualization_polydata.transform(self._plane.transformation_matrix)
                for sketch_item in self._current_sketch_context
            ]
        )
        self.plot(
            view_2d=view_2d,
            screenshot=screenshot,
            use_trame=use_trame,
            selected_pd_objects=sketches_polydata_selection,
            **plotting_options,
        )

    def sketch_polydata(self) -> List["PolyData"]:
        """
        Get polydata configuration for all objects of the sketch to the scene.

        Returns
        -------
        List[~pyvista.PolyData]
            List of the polydata configuration for all edges and faces in the sketch.
        """
        sketches_polydata = []
        sketches_polydata.extend(self.sketch_polydata_edges())
        sketches_polydata.extend(self.sketch_polydata_faces())
        return sketches_polydata

    def sketch_polydata_faces(self) -> List["PolyData"]:
        """
        Get polydata configuration for all faces of the sketch to the scene.

        Returns
        -------
        List[~pyvista.PolyData]
            List of the polydata configuration for faces in the sketch.
        """
        sketches_polydata_faces = [
            face.visualization_polydata.transform(self._plane.transformation_matrix)
            for face in self.faces
        ]
        return sketches_polydata_faces

    def sketch_polydata_edges(self) -> List["PolyData"]:
        """
        Get polydata configuration for all edges of the sketch to the scene.

        Returns
        -------
        List[~pyvista.PolyData]
            List of the polydata configuration for edges in the sketch.
        """
        sketches_polydata_edges = [
            edge.visualization_polydata.transform(self._plane.transformation_matrix)
            for edge in self.edges
        ]
        return sketches_polydata_edges



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for managing an edge."""

from enum import Enum, unique

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.edges_pb2_grpc import EdgesStub
from beartype.typing import TYPE_CHECKING, List
from pint import Quantity

from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.misc.checks import ensure_design_is_active
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.face import Face


@unique
class CurveType(Enum):
    """Provides values for the curve types supported by the Geometry service."""

    CURVETYPE_UNKNOWN = 0
    CURVETYPE_LINE = 1
    CURVETYPE_CIRCLE = 2
    CURVETYPE_ELLIPSE = 3
    CURVETYPE_NURBS = 4
    CURVETYPE_PROCEDURAL = 5


class Edge:
    """
    Represents a single edge of a body within the design assembly.

    This class synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    curve_type : CurveType
        Type of curve that the edge forms.
    body : Body
        Parent body that the edge constructs.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    """

    def __init__(self, id: str, curve_type: CurveType, body: "Body", grpc_client: GrpcClient):
        """Initialize ``Edge`` class."""
        self._id = id
        self._curve_type = curve_type
        self._body = body
        self._grpc_client = grpc_client
        self._edges_stub = EdgesStub(grpc_client.channel)

    @property
    def id(self) -> str:
        """ID of the edge."""
        return self._id

    @property
    def _grpc_id(self) -> EntityIdentifier:
        """Entity identifier of this edge on the server side."""
        return EntityIdentifier(id=self._id)

    @property
    @protect_grpc
    @ensure_design_is_active
    def length(self) -> Quantity:
        """Calculated length of the edge."""
        self._grpc_client.log.debug("Requesting edge length from server.")
        length_response = self._edges_stub.GetLength(self._grpc_id)
        return Quantity(length_response.length, DEFAULT_UNITS.SERVER_LENGTH)

    @property
    def curve_type(self) -> CurveType:
        """Curve type of the edge."""
        return self._curve_type

    @property
    @protect_grpc
    @ensure_design_is_active
    def faces(self) -> List["Face"]:
        """Faces that contain the edge."""
        from ansys.geometry.core.designer.face import Face, SurfaceType

        self._grpc_client.log.debug("Requesting edge faces from server.")
        grpc_faces = self._edges_stub.GetFaces(self._grpc_id).faces
        return [
            Face(grpc_face.id, SurfaceType(grpc_face.surface_type), self._body, self._grpc_client)
            for grpc_face in grpc_faces
        ]

    @property
    @protect_grpc
    @ensure_design_is_active
    def start_point(self) -> Point3D:
        """Edge start point."""
        self._grpc_client.log.debug("Requesting edge points from server.")
        point = self._edges_stub.GetStartAndEndPoints(self._grpc_id).start
        return Point3D([point.x, point.y, point.z])

    @property
    @protect_grpc
    @ensure_design_is_active
    def end_point(self) -> Point3D:
        """Edge end point."""
        self._grpc_client.log.debug("Requesting edge points from server.")
        point = self._edges_stub.GetStartAndEndPoints(self._grpc_id).end
        return Point3D([point.x, point.y, point.z])
